"""Code to interact with user configuration and state."""

import json
from pathlib import Path
from typing import Optional

import typer


def get_config_dir() -> Path:
    """Get the path to the configuration directory.

    :returns: The path to the configuration directory.
    """
    return Path(typer.get_app_dir("gweave"))


def get_config_file() -> Path:
    """Get the path to the configuration file.

    :returns: The path to the configuration file.
    """
    return get_config_dir() / "config.json"


def get_auth_token_file() -> Path:
    """Get the path to the authentication token file.

    :returns: The path to the authentication token file.
    """
    return get_config_dir() / "auth_token.json"


def get_auth_token() -> Optional[dict]:
    """Get the authentication token data from the authentication token file.

    :returns: The authentication token.
    """
    auth_token_file = get_auth_token_file()

    if not auth_token_file.is_file():
        return None

    with open(auth_token_file, "r") as f:
        token = json.load(f)

    return token


def save_auth_token(token: dict) -> None:
    """Save the authentication token to the authentication token file.

    :param token: The authentication token.
    """
    auth_token_file = get_auth_token_file()
    auth_token_file.parent.mkdir(parents=True, exist_ok=True)

    with open(auth_token_file, "w") as f:
        json.dump(token, f)



"""Utility functions for prompting when allowing None values."""

from typing import Any, Optional, TypeVar

import typer

NONE_REPRS = ["--", "NONE"]

T = TypeVar("T")


def value_represents_none(value: Any) -> bool:  # noqa: ANN401
    """Check if the value represents None.

    A value is considered to represent None if the upper case version of its string
    representation is in the NONE_REPRS list.

    :param value: The value to check.
    :return: True if the value represents None, False otherwise.
    """
    try:
        return str(value).upper() in NONE_REPRS
    except (TypeError, AttributeError):
        return False


def allow_none_str() -> str:
    """Format the string prompt when a field allows None values.

    :return: A string describing the allowed None values.
    """
    return f" (or {'/'.join(NONE_REPRS)} to leave blank)"


def prompt_if_none(field_name: str, value: Optional[T] = None) -> Optional[T]:
    """Prompt the user to enter a value if the value is None.

    :param value: The value to check.
    :param field_name: The name of the field to prompt for.
    :return: The value entered by the user.
    """
    if value is None:
        value = typer.prompt(
            f"Please enter a value for {field_name.capitalize()}"
        ).strip()
    return value



"""Utility functions for prompting boolean values."""

from typing import Optional, TypeVar

import typer

T = TypeVar("T")


def is_bool(field_type: T) -> bool:
    """Check if the field_type is a bool.

    :param field_type: The type to check.
    """
    return field_type is bool


def prompt_for_bool(field_name: Optional[str] = None, allow_none: bool = False) -> bool:
    """Prompt the user to enter a boolean value.

    :return: The boolean value entered by the user.
    """
    prompt_str = "Please enter a boolean value"
    if field_name is not None:
        prompt_str = f"{field_name.capitalize()}?"
    return typer.confirm(f"{prompt_str} ")



"""Utility functions for printing tables."""

from typing import List

from geneweaver.core.types import DictRow
from rich.console import Console
from rich.table import Table

console = Console()


def print_tabular_data(headers: List[str], rows: List[DictRow]) -> None:
    """Print the data in a formatted table.

    :param headers: The headers for the table.
    :param rows: The rows for the table.
    """
    table = Table(*headers)
    for row in rows:
        table.add_row(*(str(r) for r in row.values()))
    console.print(table)



"""Utility functions for use with the CLI."""

from typing import Any, Optional, Type

import typer
from geneweaver.client.utils.cli.prompt.bool import is_bool, prompt_for_bool
from geneweaver.client.utils.cli.prompt.enum import (
    is_enum_or_enum_union,
    prompt_for_enum_selection,
)
from geneweaver.client.utils.cli.prompt.generic import prompt_generic
from geneweaver.client.utils.cli.prompt.list import (
    is_list_of_str_or_int,
    prompt_for_list_selection,
)
from pydantic import BaseModel
from pydantic.fields import FieldInfo


def is_pydantic_base_model(arg: Type[Any]) -> bool:
    """Check if an argument is a pydantic base model.

    :param arg: The argument to check.
    :return: True if the argument is a pydantic base model, False otherwise.
    """
    try:
        return issubclass(arg, BaseModel)
    except TypeError:
        return False


def format_field_name(field_name: str) -> str:
    """Format a field name for display to the user.

    :param field_name: The field name to format.
    :return: The formatted field name.
    """
    return field_name.replace("_", " ").capitalize()


def prompt_to_keep_field(existing_kwargs: dict, field_name: str) -> dict:
    """Prompt the user to keep an existing field value.

    :param existing_kwargs: The existing kwargs to use as defaults.
    :param field_name: The name of the field to prompt for.
    """
    if field_name in existing_kwargs and not typer.confirm(
        f"{format_field_name(field_name)} =" f" {existing_kwargs[field_name]}, OK?",
        default=True,
    ):
        existing_kwargs.pop(field_name)
    return existing_kwargs


def prompt_for_field_by_type(
    field: FieldInfo, field_name: str, existing_kwargs: dict
) -> dict:
    """Prompt the user to enter a value for a field based on its type.

    :param field: The field to prompt for.
    :param field_name: The name of the field to prompt for.
    :param existing_kwargs: The existing kwargs to use as defaults.
    """
    field_type = field.annotation if field.annotation else str
    field_allow_none = field.default is None
    if is_pydantic_base_model(field_type):
        existing_kwargs[field_name] = prompt_for_missing_fields(
            field_type, existing_kwargs
        )
    elif is_enum_or_enum_union(field_type):
        existing_kwargs[field_name] = prompt_for_enum_selection(
            field_type, field_allow_none
        )
    elif is_list_of_str_or_int(field_type):
        existing_kwargs[field_name] = prompt_for_list_selection(
            field_type, field_allow_none
        )
    elif is_bool(field_type):
        existing_kwargs[field_name] = prompt_for_bool(field_name, field_allow_none)
    else:
        existing_kwargs[field_name] = prompt_generic(
            field_name, field_type, field_allow_none
        )
    return existing_kwargs


def prompt_for_missing_fields(
    model: Type[BaseModel],
    existing_kwargs: dict,
    exclude: Optional[set] = None,
    prompt_to_keep_existing: bool = True,
) -> dict:
    """Prompt the user to enter values for any missing fields in a pydantic model.

    :param model: The pydantic model to prompt for.
    :param existing_kwargs: The existing kwargs to use as defaults.
    :param exclude: A set of field names to exclude from prompting.
    :param prompt_to_keep_existing: Whether to prompt the user to keep existing
    :return: A dictionary of kwargs for the model.
    """
    for field_name, field_info in model.model_fields.items():
        if exclude and field_name in exclude:
            continue

        if prompt_to_keep_existing and field_name in existing_kwargs:
            prompt_to_keep_field(existing_kwargs, field_name)

        if field_name not in existing_kwargs:
            prompt_for_field_by_type(field_info, field_name, existing_kwargs)

    return existing_kwargs



"""Base class for all datasets."""

import io
import zipfile
from pathlib import Path

import pandas as pd
import requests


class BaseDataset:
    """Base class for all datasets."""

    URL = None
    DS_FOLDER = None
    UNZIPPED_LOC = None

    def __init__(self, base_folder: str = "data") -> None:
        """Initialize the BaseDataset."""
        self.base_folder = Path(base_folder)
        self.dataset_skip_rows = 0
        self._pandas_df = None
        self._pandas_read_f = pd.read_excel

    @property
    def dataset_path(self) -> Path:
        """Return the dataset's path."""
        return self.base_folder / self.DS_FOLDER / self.UNZIPPED_LOC

    def download_zip_file(self, redownload: bool = False) -> None:
        """Download the zip file from the dataset's URL."""
        if not redownload and self.dataset_path.is_file():
            return None

        response = requests.get(self.URL)
        response.raise_for_status()

        z = zipfile.ZipFile(io.BytesIO(response.content))

        z.extractall(self.base_folder / self.DS_FOLDER)

    def as_pandas(self) -> pd.DataFrame:
        """Return the dataset as a Pandas DataFrame."""
        if self._pandas_df is None:
            self.read_pandas()
        return self._pandas_df

    def read_pandas(self) -> None:
        """Read the dataset into a Pandas DataFrame."""
        self._pandas_df = self._pandas_read_f(
            self.dataset_path, skiprows=self.dataset_skip_rows
        )



"""Root of the alpha cli subcommand."""

import typer
from geneweaver.client.cli.alpha import api, datasets, geneset, parse

HELP_MESSAGE = """
These commands are in alpha testing and are considered [bold]experimental[/bold].

They may be removed or changed in the future [italic bold]without warning[/italic bold],
and have a higher risk of bugs.

Alpha commands do not have the same level of testing requirements
as the rest of the CLI, and [bold underline magenta]may break or change in unexpected \
ways[/bold underline magenta].

:warning: [bold red]Use at your own risk.[/bold red] :warning:
"""

cli = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")

cli.add_typer(geneset.cli, name="geneset", help=geneset.HELP_MESSAGE)
# Geneset aliases
cli.add_typer(geneset.cli, name="genesets", help=geneset.HELP_MESSAGE, hidden=True)
cli.add_typer(geneset.cli, name="gs", help=geneset.HELP_MESSAGE, hidden=True)


cli.add_typer(parse.cli, name="parse", help=parse.HELP_MESSAGE)
cli.add_typer(api.cli, name="api", help=api.HELP_MESSAGE)
cli.add_typer(datasets.cli, name="datasets", help=datasets.HELP_MESSAGE)



"""GeneWeaver Client configuration module."""

from typing import List, Optional

from pydantic import model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Self


class Settings(BaseSettings):
    """Settings class for GeneWeaver Client."""

    AUTH_DOMAIN: str = "thejacksonlaboratory.auth0.com"
    AUTH_CLIENT_ID: str = "f8QZPcIrPIG6DIeWR2Rr3C8X5bzx8zBz"
    AUTH_ALGORITHMS: List[str] = ["RS256"]
    AUTH_SCOPES: List[str] = ["openid", "profile", "email", "offline_access"]
    AUTH_AUDIENCE: str = "https://cube.jax.org"

    API_HOST: str = "https://geneweaver.jax.org"
    API_PATH: str = "/api"
    AON_API_PATH: str = "/aon/api"

    API_URL: Optional[str] = None
    AON_API_URL: Optional[str] = None

    GEDB: Optional[str] = None

    API_KEY: Optional[str] = None

    @model_validator(mode="after")
    def assemble_api_urls(self) -> Self:
        """Build the API URLs."""
        if not self.API_URL:
            self.API_URL = self.API_HOST + self.API_PATH
        if not self.AON_API_URL:
            self.AON_API_URL = self.API_HOST + self.AON_API_PATH
        if not self.GEDB:
            self.GEDB = self.API_HOST + "/gedb"
        return self

    model_config = SettingsConfigDict(
        env_file=".env",
        case_sensitive=False,
        extra="ignore",
    )



"""Print information about a file."""

from pathlib import Path
from typing import List


def print_metadata_xlsx(
    file_path: Path,
    sheet_names: List[str],
    sheet_metadata: List[List[str]],
    n_sheets: int,
) -> None:
    """Print metadata from an Excel file.

    :param file_path: The file path to the Excel file.
    :param sheet_names: A list of sheet names in the Excel file.
    :param sheet_metadata: A list of metadata for each sheet in the Excel file.
    :param n_sheets: The number of sheets in the Excel file.
    """
    print(f"{file_path} contains {n_sheets} sheets:")

    for name, metadata in zip(sheet_names, sheet_metadata):  # noqa: B905
        print(f" - {name} - {metadata}")


def print_metadata_csv(file_path: Path, metadata: List[str]) -> None:
    """Print the metadata from a CSV file.

    :param file_path: The file path to the CSV file.
    :param metadata: A list of metadata for the CSV file.
    """
    print(f"{file_path} contains:")
    for metadata_line in metadata:
        print(f" - {metadata_line}")
